/* The code here is based upon work at AT&T by Steven Fortune (1994).
 * The convex hull code was adapted from code by Ken Clarkson (1996).
 *
 * Copyright (c) 1994, 1996 by AT&T Bell Laboratories
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>

#include <slang.h>
#define MYALLOC (void *)SLmalloc
#define MYFREE(x) SLfree((char *)(x))

#include "voronoi.h"

typedef struct tagPoint
{
   double x ;
   double y ;
} Point ;

/* structure used both for sites and for vertices */

typedef struct tagSite
{
   Point coord ;
   int sitenbr ;
   int refcnt ;
} Site ;

typedef struct tagEdge
{
   double a, b, c ;
   Site * ep[2] ;
   Site * reg[2] ;
   int edgenbr ;
} Edge ;

#define DELETED_EDGE ((Edge *)-2)
#define LEFT_EDGE 0
#define RIGHT_EDGE 1

typedef struct tagHalfedge
{
   struct tagHalfedge * ELleft ;
   struct tagHalfedge * ELright ;
   Edge * ELedge ;
   int ELrefcnt ;
   char ELpm ;
   Site * vertex ;
   double ystar ;
   struct tagHalfedge * PQnext ;
} Halfedge ;

#define DEREF_SITE(v,s) \
   { \
      if ((s) != NULL) \
	{ \
	   if ((s)->refcnt > 1) \
	     { \
		(s)->refcnt--; \
	     } \
	   else \
	     { \
		(s)->refcnt = 0; \
		sweep_makefree_site ((v), (s)); \
	     } \
	} \
   }
#define REF_SITE(s) (s)->refcnt++

typedef struct Free_Site_List_Type Free_Site_List_Type;
typedef struct Free_Edge_List_Type Free_Edge_List_Type;
typedef struct Free_HalfEdge_List_Type Free_HalfEdge_List_Type;

struct _Voronoi_Type
{
   Free_Site_List_Type *sfl;
   double *xpts, *ypts;
   unsigned int *sort_indices;
   unsigned int site_idx;
   unsigned int nsites;
   double xmin, xmax, ymin, ymax;
   double deltax, deltay;

   Free_Edge_List_Type *efl;
   unsigned int nedges, nvertices, sqrt_nsites;

   Free_HalfEdge_List_Type *hfl;

   Site *bottomsite;
   int el_hashsize;
   Halfedge *el_leftend, *el_rightend;
   Halfedge **el_hash;

   int pq_min, pq_count, pq_hashsize;
   Halfedge *pq_hash;

   int triangulate;
   int debug;

   /* Callbacks */
   int (*out_vertex)(Voronoi_Type *, Site *);
   void *out_vertex_data;

   int (*out_ep) (Voronoi_Type *, Edge *);
   void *out_ep_data;

   int (*out_bisector) (Voronoi_Type *, Edge *);
   void *out_bisector_data;

   int (*out_site) (Voronoi_Type *, Site *);
   void *out_site_data;

   int (*out_triple) (Voronoi_Type *, Site *, Site *, Site *);
   void *out_triple_data;
};

#ifndef MYFREE
# define MYFREE myfree
static void myfree (void *p)
{
   if (p != NULL)
     free (p);
}
#endif

#ifndef MYALLOC
# define MYALLOC myalloc
static void *myalloc(size_t n)
{
   void *t;

   if (NULL == ( t = malloc(n)))
     fprintf(stderr,"Out of memory\n") ;

   return t;
}
#endif

typedef struct tagFreenode
{
   struct tagFreenode *nextfree;
}
Freenode;

#define ALLOCATE_SINGLE_OBJECTS 0
#define PRINT_MALLOC_STATS	0

/* This work-around tracks prevents frees any residual allocated memory
 * at voronoi_close.
 */
typedef struct Allocated_Block_Type
{
   Freenode *block;
   struct Allocated_Block_Type *next;
}
Allocated_Block_Type;

#define FREELIST_STRUCT_FIELDS \
   Allocated_Block_Type *memory_blocks; \
   Freenode *head; \
   unsigned int nodesize;

typedef struct tagFreelist
{
   FREELIST_STRUCT_FIELDS
} Freelist ;

struct Free_Site_List_Type
{
   FREELIST_STRUCT_FIELDS
};

struct Free_Edge_List_Type
{
   FREELIST_STRUCT_FIELDS
};

struct Free_HalfEdge_List_Type
{
   FREELIST_STRUCT_FIELDS
};

static int freeinit (Freelist * fl, size_t size)
{
   fl->memory_blocks = NULL;
   fl->head = NULL;
   fl->nodesize = size ;
   return 0;
}

static void free_list (Freelist *fl)
{
   Allocated_Block_Type *b;

   if (fl == NULL)
     return;

   b = fl->memory_blocks;
   while (b != NULL)
     {
	Allocated_Block_Type *bnext = b->next;
	MYFREE (b->block);
	MYFREE (b);
	b = bnext;
     }
   MYFREE (fl);
}

static void makefree (Freelist *fl, Freenode * curr)
{
#if ALLOCATE_SINGLE_OBJECTS
   MYFREE (curr);
#else
   curr->nextfree = fl->head ;
   fl->head = curr ;
#endif
}

static int alloc_free_site_list (Voronoi_Type *v)
{
   if (NULL == (v->sfl = (Free_Site_List_Type *)MYALLOC (sizeof (Free_Site_List_Type))))
     return -1;

   return freeinit ((Freelist *)v->sfl, sizeof(Site));
}

static int alloc_free_edge_list (Voronoi_Type *v)
{
   if (NULL == (v->efl = (Free_Edge_List_Type *)MYALLOC (sizeof (Free_Edge_List_Type))))
     return -1;

   return freeinit ((Freelist *)v->efl, sizeof(Edge));
}

static int alloc_free_halfedge_list (Voronoi_Type *v)
{
   if (NULL == (v->hfl = (Free_HalfEdge_List_Type *)MYALLOC (sizeof (Free_HalfEdge_List_Type))))
     return -1;
   return freeinit ((Freelist *)v->hfl, sizeof(Halfedge));
}

static void free_site_list (Free_Site_List_Type *sfl)
{
   free_list ((Freelist *) sfl);
}
static void free_edge_list (Free_Edge_List_Type *efl)
{
   free_list ((Freelist *) efl);
}
static void free_halfedge_list (Free_HalfEdge_List_Type *hfl)
{
   free_list ((Freelist *) hfl);
}

static Freenode *
getfree_node (Voronoi_Type *v, Freelist * fl)
{
#if ALLOCATE_SINGLE_OBJECTS
   return (Freenode *) MYALLOC (fl->nodesize);
#else
   Freenode * t ;
   if (fl->head == (Freenode *)NULL)
     {
	unsigned int i;
	unsigned int sqrt_nsites = v->sqrt_nsites;
	Allocated_Block_Type *b;

	t = (Freenode *) MYALLOC(sqrt_nsites * fl->nodesize) ;
	if (t == NULL)
	  return NULL;

	b = (Allocated_Block_Type *) MYALLOC (sizeof (Allocated_Block_Type));
	if (b == NULL)
	  {
	     MYFREE (t);
	     return NULL;
	  }
	b->block = t;
	b->next = fl->memory_blocks;
	fl->memory_blocks = b;

	if (t == NULL)
	  return NULL;

        for (i = 0 ; i < sqrt_nsites ; i++)
	  {
	     makefree(fl, (Freenode *)((char *)t+i*fl->nodesize)) ;
	  }
     }
   t = fl->head ;
   fl->head = (fl->head)->nextfree ;
   return (t) ;
#endif
}

#if PRINT_MALLOC_STATS
static unsigned int Num_Sites_Allocated;
static unsigned int Max_Num_Sites_Allocated;
static unsigned int Num_Edges_Allocated;
static unsigned int Max_Num_Edges_Allocated;
static unsigned int Num_HalfEdges_Allocated;
static unsigned int Max_Num_HalfEdges_Allocated;
#endif
static Site *sweep_getfree_site (Voronoi_Type *v)
{
#if PRINT_MALLOC_STATS
   Num_Sites_Allocated++;
   if (Num_Sites_Allocated > Max_Num_Sites_Allocated)
     Max_Num_Sites_Allocated = Num_Sites_Allocated;
#endif
   return (Site *) getfree_node (v, (Freelist *)v->sfl);
}

static Halfedge *sweep_getfree_halfedge (Voronoi_Type *v)
{
#if PRINT_MALLOC_STATS
   Num_HalfEdges_Allocated++;
   if (Num_HalfEdges_Allocated > Max_Num_HalfEdges_Allocated)
     Max_Num_HalfEdges_Allocated = Num_HalfEdges_Allocated;
#endif
   return (Halfedge *) getfree_node (v, (Freelist *)v->hfl);
}

static void sweep_makefree_site (Voronoi_Type *v, Site *s)
{
#if PRINT_MALLOC_STATS
   Num_Sites_Allocated--;
#endif
   makefree ((Freelist *)v->sfl, (Freenode *)s);
}

static Edge *sweep_getfree_edge (Voronoi_Type *v)
{
#if PRINT_MALLOC_STATS
   Num_Edges_Allocated++;
   if (Num_Edges_Allocated > Max_Num_Edges_Allocated)
     Max_Num_Edges_Allocated = Num_Edges_Allocated;
#endif
   return (Edge *) getfree_node (v, (Freelist *)v->efl);
}

static void sweep_makefree_halfedge (Voronoi_Type *v, Halfedge *he)
{
#if PRINT_MALLOC_STATS
   Num_HalfEdges_Allocated--;
#endif
   if (he->vertex != NULL)	       /* JED */
     DEREF_SITE (v, he->vertex);

   makefree ((Freelist *)v->hfl, (Freenode *)he);
}

static void sweep_makefree_edge (Voronoi_Type *v, Edge *e)
{
#if PRINT_MALLOC_STATS
   Num_Edges_Allocated--;
#endif
   makefree ((Freelist *)v->efl, (Freenode *)e);
}

static Edge *bisect(Voronoi_Type *v, Site * s1, Site * s2)
{
   double dx, dy, adx, ady ;
   Edge * newedge ;

   newedge = sweep_getfree_edge (v) ;
   if (newedge == NULL)
     return NULL;

   newedge->reg[0] = s1 ;
   newedge->reg[1] = s2 ;
   REF_SITE(s1) ;
   REF_SITE(s2) ;
   newedge->ep[0] = newedge->ep[1] = (Site *)NULL ;
   dx = s2->coord.x - s1->coord.x ;
   dy = s2->coord.y - s1->coord.y ;
   adx = dx>0 ? dx : -dx ;
   ady = dy>0 ? dy : -dy ;
   newedge->c = s1->coord.x * dx + s1->coord.y * dy + (dx*dx +
						       dy*dy) * 0.5 ;
   if (adx > ady)
     {
        newedge->a = 1.0 ;
        newedge->b = dy/dx ;
        newedge->c /= dx ;
     }
   else
     {
        newedge->b = 1.0 ;
        newedge->a = dx/dy ;
        newedge->c /= dy ;
     }
   newedge->edgenbr = v->nedges ;

   if (-1 == (*v->out_bisector) (v, newedge))
     return NULL;

   v->nedges++ ;
   return (newedge) ;
}

static int intersect(Voronoi_Type *v, Halfedge * el1, Halfedge * el2, Site **sp)
{
   Edge * e1, * e2, * e ;
   Halfedge * el ;
   double d, xint, yint ;
   int right_of_site ;
   Site * s ;

   e1 = el1->ELedge ;
   e2 = el2->ELedge ;
   if ((e1 == (Edge*)NULL) || (e2 == (Edge*)NULL))
     {
	*sp = NULL;
        return 0;
     }
   if (e1->reg[1] == e2->reg[1])
     {
	*sp = NULL;
        return 0;
     }
   d = (e1->a * e2->b) - (e1->b * e2->a) ;
   if ((-1.0e-10 < d) && (d < 1.0e-10))
     {
	*sp = NULL;
        return 0;
     }
   xint = (e1->c * e2->b - e2->c * e1->b) / d ;
   yint = (e2->c * e1->a - e1->c * e2->a) / d ;
   if ((e1->reg[1]->coord.y < e2->reg[1]->coord.y) ||
       (e1->reg[1]->coord.y == e2->reg[1]->coord.y &&
	   e1->reg[1]->coord.x < e2->reg[1]->coord.x))
     {
        el = el1 ;
        e = e1 ;
     }
   else
     {
        el = el2 ;
        e = e2 ;
     }
   right_of_site = (xint >= e->reg[1]->coord.x) ;
   if ((right_of_site && (el->ELpm == LEFT_EDGE)) ||
       (!right_of_site && (el->ELpm == RIGHT_EDGE)))
     {
	*sp = NULL;
        return 0;
     }
   s = sweep_getfree_site (v) ;
   if (s == NULL)
     {
	*sp = NULL;
	return -1;
     }

   s->refcnt = 1 ;		       /* JED */
   s->coord.x = xint ;
   s->coord.y = yint ;
   *sp = s;
   return 1;
}

/*** returns 1 if p is to right of halfedge e ***/

static int right_of(Halfedge * el, Point * p)
{
   Edge * e ;
   Site * topsite ;
   int right_of_site, above, fast ;
   double dxp, dyp, dxs, t1, t2, t3, yl ;

   e = el->ELedge ;
   topsite = e->reg[1] ;
   right_of_site = (p->x > topsite->coord.x) ;
   if (right_of_site && (el->ELpm == LEFT_EDGE))
     {
        return (1) ;
     }
   if(!right_of_site && (el->ELpm == RIGHT_EDGE))
     {
        return (0) ;
     }
   if (e->a == 1.0)
     {
        dyp = p->y - topsite->coord.y ;
        dxp = p->x - topsite->coord.x ;
        fast = 0 ;
        if ((!right_of_site & (e->b < 0.0)) ||
	    (right_of_site & (e->b >= 0.0)))
	  {
	     fast = above = (dyp >= e->b*dxp) ;
	  }
        else
	  {
	     above = ((p->x + p->y * e->b) > (e->c)) ;
	     if (e->b < 0.0)
	       {
		  above = !above ;
	       }
	     if (!above)
	       {
		  fast = 1 ;
	       }
	  }
        if (!fast)
	  {
	     dxs = topsite->coord.x - (e->reg[0])->coord.x ;
	     above = (e->b * (dxp*dxp - dyp*dyp))
	       <
	       (dxs * dyp * (1.0 + 2.0 * dxp /
			     dxs + e->b * e->b)) ;
	     if (e->b < 0.0)
	       {
		  above = !above ;
	       }
	  }
     }
   else  /*** e->b == 1.0 ***/
     {
        yl = e->c - e->a * p->x ;
        t1 = p->y - yl ;
        t2 = p->x - topsite->coord.x ;
        t3 = yl - topsite->coord.y ;
        above = ((t1*t1) > ((t2 * t2) + (t3 * t3))) ;
     }
   return (el->ELpm == LEFT_EDGE ? above : !above) ;
}

static int endpoint (Voronoi_Type *v, Edge * e, int lr, Site * s)
{
   if (e->ep[lr] != NULL) DEREF_SITE (v, e->ep[lr]);  /* JED */
   e->ep[lr] = s ;
   REF_SITE(s) ;
   if (e->ep[RIGHT_EDGE-lr] == (Site *)NULL)
     {
        return 0;
     }

   if (-1 == (*v->out_ep) (v, e))
     return -1;

   if (e->ep[0] != NULL) DEREF_SITE (v, e->ep[0]);   /* JED */
   if (e->ep[1] != NULL) DEREF_SITE (v, e->ep[1]);   /* JED */

   DEREF_SITE(v, e->reg[LEFT_EDGE]) ;
   DEREF_SITE(v, e->reg[RIGHT_EDGE]) ;
   e->reg[LEFT_EDGE] = NULL;	       /* JED */
   e->reg[RIGHT_EDGE] = NULL;	       /* JED */
   sweep_makefree_edge (v, e) ;
   return 0;
}

static double dist(Site * s, Site * t)
{
   return hypot (s->coord.x - t->coord.x, s->coord.y - t->coord.y);
}

static Halfedge *HEcreate(Voronoi_Type *v, Edge * e, int pm)
{
   Halfedge *he ;

   he = sweep_getfree_halfedge (v) ;
   if (he == NULL)
     return NULL;

   memset ((char *)he, 0, sizeof (Halfedge));
   he->ELedge = e ;
   he->ELpm = pm ;
   return he;
}

static void ELinsert(Halfedge * lb, Halfedge * he)
{
   he->ELrefcnt++;		       /* JED */

   he->ELleft = lb ;
   he->ELright = lb->ELright ;
   (lb->ELright)->ELleft = he ;
   lb->ELright = he ;
}

/* Get entry from hash table, pruning any deleted nodes */

static Halfedge *ELgethash(Voronoi_Type *v, int b)
{
   Halfedge * he ;

   if ((b < 0) || (b >= v->el_hashsize))
     {
        return ((Halfedge *)NULL) ;
     }
   he = v->el_hash[b] ;
   if ((he == (Halfedge *)NULL) || (he->ELedge != DELETED_EDGE))
     {
        return (he) ;
     }
   /* Hash table points to deleted half edge.  Patch as necessary. */
   v->el_hash[b] = (Halfedge *)NULL ;
   he->ELrefcnt--;
   if (he->ELrefcnt == 0)
     {
        sweep_makefree_halfedge (v, he);
     }
   return ((Halfedge *)NULL) ;
}

static Halfedge *ELleftbnd (Voronoi_Type *v, Point * p)
{
   int i, bucket ;
   Halfedge * he ;

   /* Use hash table to get close to desired halfedge */
   bucket = (p->x - v->xmin) / v->deltax * v->el_hashsize ;
   if (bucket < 0)
     {
        bucket = 0 ;
     }
   if (bucket >= v->el_hashsize)
     {
        bucket = v->el_hashsize - 1 ;
     }
   he = ELgethash(v, bucket) ;
   if (he == (Halfedge *)NULL)
     {
	i = 0;
	while (1)
	  {
	     i++;
	     if ((he = ELgethash(v, bucket-i)) != (Halfedge *)NULL)
	       {
		  break ;
	       }
	     if ((he = ELgethash(v, bucket+i)) != (Halfedge *)NULL)
	       {
		  break ;
	       }
	  }
     }
   /* Now search linear list of halfedges for the corect one */
   if (he == v->el_leftend || (he != v->el_rightend && right_of(he,p)))
     {
        do
	  {
	     he = he->ELright ;
	  }
	while (he != v->el_rightend && right_of(he,p)) ;
        he = he->ELleft ;
     }
   else
     {
        do
	  {
	     he = he->ELleft ;
	  }
	while (he != v->el_leftend && !right_of(he,p)) ;
     }
   /*** Update hash table and reference counts ***/
   if ((bucket > 0) && (bucket < v->el_hashsize-1))
     {
        if (v->el_hash[bucket] != (Halfedge *)NULL)
	  {
	     (v->el_hash[bucket]->ELrefcnt)-- ;
	  }
        v->el_hash[bucket] = he ;
	he->ELrefcnt++;
     }
   return (he) ;
}

/*** This delete routine can't reclaim node, since pointers from hash
 : table may be present.
 ***/

static void ELdelete(Voronoi_Type *v, Halfedge * he)
{
   he->ELrefcnt--;		       /* JED */

   if (he->ELleft != NULL)
     (he->ELleft)->ELright = he->ELright ;
   if (he->ELright != NULL)
     (he->ELright)->ELleft = he->ELleft ;

   he->ELedge = DELETED_EDGE ;

   if (he->ELrefcnt == 0)		       /* JED */
     sweep_makefree_halfedge (v, he);
}

static Halfedge *ELright(Halfedge * he)
{
   return (he->ELright) ;
}

static Halfedge *ELleft(Halfedge * he)
{
   return (he->ELleft) ;
}

static Site *leftreg(Voronoi_Type *v, Halfedge * he)
{
   if (he->ELedge == (Edge *)NULL)
     {
        return(v->bottomsite) ;
     }
   return (he->ELpm == LEFT_EDGE ? he->ELedge->reg[LEFT_EDGE] :
	   he->ELedge->reg[RIGHT_EDGE]) ;
}

static Site *rightreg(Voronoi_Type *v, Halfedge * he)
{
   if (he->ELedge == (Edge *)NULL)
     {
        return(v->bottomsite) ;
     }
   return (he->ELpm == LEFT_EDGE ? he->ELedge->reg[RIGHT_EDGE] :
	   he->ELedge->reg[LEFT_EDGE]) ;
}

static int PQbucket (Voronoi_Type *v, Halfedge * he)
{
   int bucket ;

   if (he->ystar < v->ymin)
     bucket = 0;
   else if (he->ystar >= v->ymax)
     bucket = v->pq_hashsize-1;
   else
     bucket = (he->ystar - v->ymin)/v->deltay * v->pq_hashsize;

   if (bucket < 0)
     {
        bucket = 0 ;
     }
   if (bucket >= v->pq_hashsize)
     {
        bucket = v->pq_hashsize-1 ;
     }
   if (bucket < v->pq_min)
     {
        v->pq_min = bucket ;
     }
   return (bucket);
}

static int PQempty(Voronoi_Type *v)
{
   return (v->pq_count == 0) ;
}

static Point PQ_min (Voronoi_Type *v)
{
   Point answer ;

   while (v->pq_hash[v->pq_min].PQnext == (Halfedge *)NULL)
     {
        v->pq_min++ ;
     }
   answer.x = v->pq_hash[v->pq_min].PQnext->vertex->coord.x ;
   answer.y = v->pq_hash[v->pq_min].PQnext->ystar ;
   return (answer) ;
}

static Halfedge *PQextractmin (Voronoi_Type *v)
{
   Halfedge * curr ;

   curr = v->pq_hash[v->pq_min].PQnext ;
   v->pq_hash[v->pq_min].PQnext = curr->PQnext ;
   v->pq_count-- ;
   return (curr) ;
}

static void PQinsert(Voronoi_Type *v, Halfedge * he, Site * s, double offset)
{
   Halfedge * last, * next ;

   if (he->vertex != NULL) DEREF_SITE (v, he->vertex);   /* JED */
   he->vertex = s ;
   REF_SITE(s) ;
   he->ystar = s->coord.y + offset ;
   last = v->pq_hash + PQbucket(v, he);
   while ((next = last->PQnext) != (Halfedge *)NULL &&
	  (he->ystar  > next->ystar  ||
	      (he->ystar == next->ystar &&
		  s->coord.x > next->vertex->coord.x)))
     {
        last = next ;
     }
   he->PQnext = last->PQnext ;
   last->PQnext = he ;
   v->pq_count++;
}

static void PQdelete(Voronoi_Type *v, Halfedge * he)
{
   Halfedge * last;

   if (he->vertex == NULL)
     return;

   last = v->pq_hash + PQbucket(v, he);
   while (last -> PQnext != he)
     {
	last = last->PQnext ;
     }
   last->PQnext = he->PQnext;
   v->pq_count-- ;
   DEREF_SITE(v, he->vertex) ;
   he->vertex = (Site *)NULL ;
}

static int PQinitialize (Voronoi_Type *v)
{
   int hashsize;

   hashsize = 4 * v->sqrt_nsites;
   v->pq_hash = (Halfedge *)MYALLOC(hashsize * sizeof(Halfedge));
   if (v->pq_hash == NULL)
     return -1;
   memset ((char *)v->pq_hash, 0, hashsize * sizeof (Halfedge));

   v->pq_count = v->pq_min = 0 ;
   v->pq_hashsize = hashsize;
   return 0;
}

static void PQfreehash (Voronoi_Type *v)
{
   unsigned int i, hashsize = v->pq_hashsize;

   if (v->pq_hash == NULL)
     return;

   for (i = 0; i < hashsize; i++)
     {
	Halfedge *he;

	he = &v->pq_hash[i];
	while (he != NULL)
	  {
	     Halfedge *next =  he->PQnext;
	     if (he->vertex != NULL)
	       {
		  DEREF_SITE (v, he->vertex);
		  he->vertex = NULL;
	       }
	     he = next;
	  }
     }
   MYFREE (v->pq_hash);
}

static int nextsite (Voronoi_Type *v, Site **sp)
{
   Site *s;
   unsigned int i;

   if (v->site_idx >= v->nsites)
     {
	*sp = NULL;
	return 0;
     }

   if (NULL == (s = sweep_getfree_site (v)))
     {
	*sp = NULL;
	return -1;
     }

   i = v->sort_indices[v->site_idx];
   v->site_idx++;
   s->coord.x = v->xpts[i];
   s->coord.y = v->ypts[i];
   s->sitenbr = i;
   s->refcnt = 1;

   *sp = s;
   return 1;
}

static int makevertex (Voronoi_Type *v, Site * s)
{
   s->sitenbr = v->nvertices++ ;
   return (*v->out_vertex)(v, s) ;
}

static int voronoi(Voronoi_Type *v, int triangulate)
{
   Site * newsite, * bot, * top, * p, * s ;
   Point newintstar ;
   int pm ;
   Halfedge * lbnd, * rbnd, * llbnd, * rrbnd, * bisector ;
   Halfedge *el_rightend;
   Edge * e ;
   int status;

   v->triangulate = triangulate;

   if (-1 == nextsite (v, &v->bottomsite))
     return -1;

   if (-1 == (*v->out_site)(v, v->bottomsite))
     return -1;

   if (-1 == nextsite (v, &newsite))
     return -1;

   while (1)
     {
        if(!PQempty(v))
	  {
	     newintstar = PQ_min(v) ;
	  }

        if ((newsite != NULL)
	    && (PQempty(v)
		|| newsite -> coord.y < newintstar.y
		|| (newsite->coord.y == newintstar.y
		    && newsite->coord.x < newintstar.x)))
	  {
	     /* new site is smallest */
	       {
		  if (-1 == (*v->out_site)(v, newsite))
		    return -1;
	       }
	     lbnd = ELleftbnd(v, &(newsite->coord)) ;
	     rbnd = ELright(lbnd) ;
	     bot = rightreg(v, lbnd) ;

	     if (NULL == (e = bisect(v, bot, newsite)))
	       return -1;

	     if (NULL == (bisector = HEcreate(v, e, LEFT_EDGE)))
	       goto return_error;

	     ELinsert(lbnd, bisector) ;
	     status = intersect(v, lbnd, bisector, &p) ;
	     if (status > 0)
	       {
		  PQdelete(v, lbnd) ;
		  PQinsert(v, lbnd, p, dist(p,newsite)) ;
	       }
	     else if (status == -1)
	       return -1;

	     lbnd = bisector ;
	     if (NULL == (bisector = HEcreate(v, e, RIGHT_EDGE)))
	       goto return_error;

	     ELinsert(lbnd, bisector) ;
	     DEREF_SITE (v, p);	       /* JED */
	     status = intersect(v, bisector, rbnd, &p) ;
	     if (status == 1)
	       {
		  PQinsert(v, bisector, p, dist(p,newsite)) ;
	       }
	     else if (status == -1)
	       return -1;

	     DEREF_SITE (v, p);		       /* JED */
	     DEREF_SITE (v, newsite);       /* JED */
	     if (-1 == nextsite (v, &newsite))
	       return -1;

	     continue;
	  }

	if (PQempty(v))
	  break;

	/* intersection is smallest */
	lbnd = PQextractmin(v) ;
	llbnd = ELleft(lbnd) ;
	rbnd = ELright(lbnd) ;
	rrbnd = ELright(rbnd) ;
	bot = leftreg(v, lbnd) ;
	REF_SITE (bot);		       /* JED */
	top = rightreg(v, rbnd) ;
	REF_SITE (top);		       /* JED */
	if (-1 == (*v->out_triple)(v, bot, top, rightreg(v, lbnd)))
	  return -1;
	s = lbnd->vertex ;
	REF_SITE (s);		       /* JED */
	if (-1 == makevertex(v, s))
	  return -1;
	if (-1 == endpoint(v, lbnd->ELedge, lbnd->ELpm, s))
	  return -1;
	if (-1 == endpoint(v, rbnd->ELedge, rbnd->ELpm, s))
	  return -1;

	ELdelete(v, lbnd) ;
	PQdelete(v, rbnd) ;
	ELdelete(v, rbnd) ;
	pm = LEFT_EDGE ;
	if (bot->coord.y > top->coord.y)
	  {
	     Site *temp = bot ;
	     bot = top ;
	     top = temp ;
	     pm = RIGHT_EDGE ;
	  }
	e = bisect(v, bot, top) ;
	if (e == NULL)
	  return -1;

	if (NULL == (bisector = HEcreate(v, e, pm)))
	  goto return_error;
	ELinsert(llbnd, bisector) ;
	endpoint(v, e, RIGHT_EDGE-pm, s) ;
	DEREF_SITE(v, s) ;

	status = intersect(v, llbnd, bisector, &p) ;
	if (status == 1)
	  {
	     PQdelete(v, llbnd) ;
	     PQinsert(v, llbnd, p, dist(p,bot)) ;
	  }
	else if (status == -1)
	  return -1;
	DEREF_SITE (v, p);		       /* JED */
	status = intersect (v, bisector, rrbnd, &p);
	if (status == 1)
	  {
	     PQinsert(v, bisector, p, dist(p,bot));
	  }
	else if (status == -1)
	  return -1;
	DEREF_SITE (v, top);		       /* JED */
	DEREF_SITE (v, bot);		       /* JED */
	DEREF_SITE (v, p);		       /* JED */
     }

   lbnd = ELright (v->el_leftend);
   el_rightend = v->el_rightend;
   while (lbnd != el_rightend)
     {
        e = lbnd->ELedge ;

        if (-1 == (*v->out_ep)(v, e))
	  return -1;

	lbnd = ELright (lbnd);
     }

   return 0;

return_error:
   return -1;
}

/*** sort sites on y, then x, coord ***/

static Voronoi_Type *This_V;
static int scomp(const void * vs1, const void * vs2)
{
   unsigned int i, j;

   i = *(unsigned int *) vs1;
   j = *(unsigned int *) vs2;

   if (This_V->ypts[i] < This_V->ypts[j])
     return -1;
   if (This_V->ypts[i] > This_V->ypts[j])
     return 1;
   if (This_V->xpts[i] < This_V->xpts[j])
     return -1;
   if (This_V->xpts[i] > This_V->xpts[j])
     return 1;

   return 0;
}

static int create_sites (Voronoi_Type *v)
{
   unsigned int i;
   double xmin, xmax;
   unsigned int *sort_indices;
   double *x, *y;
   unsigned int num_sites;

   x = v->xpts;
   y = v->ypts;
   num_sites = v->nsites;

   if (NULL == (sort_indices = (unsigned int *)MYALLOC (num_sites * sizeof(unsigned int))))
     return -1;
   v->sort_indices = sort_indices;
   for (i = 0; i < num_sites; i++)
     sort_indices[i] = i;

   This_V = v;
   qsort((void *)v->sort_indices, num_sites, sizeof(unsigned int), scomp) ;
   This_V = NULL;

   xmin = xmax = x[sort_indices[0]];

   for (i = 1 ; i < num_sites ; ++i)
     {
	double xx = x[sort_indices[i]];
        if (xx < xmin)
	  {
	     xmin = xx;
	  }
        if (xx > xmax)
	  {
	     xmax = xx;
	  }
     }
   v->ymin = y[sort_indices[0]];
   v->ymax = y[sort_indices[num_sites-1]];
   v->xmin = xmin;
   v->xmax = xmax;
   v->sqrt_nsites = (unsigned int) sqrt(num_sites+4) ;
   v->deltay = v->ymax - v->ymin;
   v->deltax = v->xmax - v->xmin;
   return 0;
}

static int el_initialize (Voronoi_Type *v)
{
   unsigned int el_hashsize;

   if (-1 == alloc_free_halfedge_list (v))
     return -1;

   v->el_hashsize = el_hashsize = 2 * v->sqrt_nsites;
   if (NULL == (v->el_hash = (Halfedge **)MYALLOC(sizeof(Halfedge *) * el_hashsize)))
     return -1;

   memset ((char *)v->el_hash, 0, sizeof(Halfedge*) * el_hashsize);

   if (NULL == (v->el_leftend = HEcreate (v, NULL, 0)))
     return -1;

   if (NULL == (v->el_rightend = HEcreate(v, NULL, 0)))
     return -1;

   /* Note: A circular reference count gets created here */
   v->el_leftend->ELleft = (Halfedge *)NULL ;
   v->el_leftend->ELright = v->el_rightend ;
   v->el_rightend->ELleft = v->el_leftend ;
   v->el_rightend->ELright = (Halfedge *)NULL ;
   v->el_hash[0] = v->el_leftend ;
   v->el_hash[el_hashsize-1] = v->el_rightend ;
   return 0;
}

void voronoi_close (Voronoi_Type *v)
{
   if (v == NULL)
     return;
#if PRINT_MALLOC_STATS
   fprintf (stderr, "At close:\n");
   fprintf (stderr, " num sites: %u/%u\n", Num_Sites_Allocated, Max_Num_Sites_Allocated);
   fprintf (stderr, " num edges: %u/%u\n", Num_Edges_Allocated, Max_Num_Edges_Allocated);
   fprintf (stderr, " num halfedges: %u/%u\n", Num_HalfEdges_Allocated, Max_Num_HalfEdges_Allocated);
#endif
   PQfreehash (v);

   free_site_list (v->sfl);
   free_edge_list (v->efl);
   free_halfedge_list (v->hfl);

   MYFREE (v->el_hash);
   MYFREE (v->sort_indices);
   MYFREE (v);

#if PRINT_MALLOC_STATS
   fprintf (stderr, "After close:\n");
   fprintf (stderr, " num sites: %u\n", Num_Sites_Allocated);
   fprintf (stderr, " num edges: %u\n", Num_Edges_Allocated);
   fprintf (stderr, " num halfedges: %u\n", Num_HalfEdges_Allocated);
#endif
}

static int def_out_bisector_callback (Voronoi_Type *v, Edge *e)
{
   if (v->triangulate)
     return 0;

   if (v->debug)
     {
        printf("line(%d) %gx+%gy=%g, bisecting %d %d\n", e->edgenbr,
	       e->a, e->b, e->c, e->reg[LEFT_EDGE]->sitenbr,
	       e->reg[RIGHT_EDGE]->sitenbr) ;
     }

   printf("l %f %f %f\n", e->a, e->b, e->c) ;
   return 0;
}

static int def_out_ep_callback (Voronoi_Type *v, Edge * e)
{
   if (v->triangulate)
     return 0;

   printf("e %d", e->edgenbr);
   printf(" %d ", e->ep[LEFT_EDGE] != (Site *)NULL
	  ? e->ep[LEFT_EDGE]->sitenbr : -1) ;
   printf("%d\n", e->ep[RIGHT_EDGE] != (Site *)NULL
	  ? e->ep[RIGHT_EDGE]->sitenbr : -1) ;
   return 0;
}

static int def_out_vertex_callback (Voronoi_Type *v, Site *s)
{
   if (v->triangulate)
     return 0;

   if (v->debug)
     {
        printf("vertex(%d) at %f %f\n", s->sitenbr, s->coord.x, s->coord.y) ;
     }
   printf ("v %f %f\n", s->coord.x, s->coord.y) ;
   return 0;
}

static int def_out_site_callback (Voronoi_Type *v, Site * s)
{
   if (v->triangulate)
     return 0;

   if (v->debug)
     {
        printf("site (%d) at %f %f\n", s->sitenbr, s->coord.x, s->coord.y) ;
     }
   printf("s %f %f\n", s->coord.x, s->coord.y) ;
   return 0;
}

static int def_out_triple_callback (Voronoi_Type *v, Site * s1, Site * s2, Site * s3)
{
   if (v->triangulate == 0)
     return 0;

   printf ("%d %d %d\n", s1->sitenbr, s2->sitenbr, s3->sitenbr);
   return 0;
}

Voronoi_Type *voronoi_open (double *x, double *y, unsigned int npts)
{
   Voronoi_Type *v;

   if (NULL == (v = (Voronoi_Type *)MYALLOC (sizeof (Voronoi_Type))))
     return NULL;

   memset ((char *)v, 0, sizeof (Voronoi_Type));

   if (-1 == alloc_free_site_list (v))
     {
	voronoi_close (v);
	return NULL;
     }
   if (-1 == alloc_free_edge_list (v))
     {
	voronoi_close (v);
	return NULL;
     }
   v->xpts = x;
   v->ypts = y;
   v->nsites = npts;

   v->out_vertex = def_out_vertex_callback;
   v->out_ep = def_out_ep_callback;
   v->out_bisector = def_out_bisector_callback;
   v->out_site = def_out_site_callback;
   v->out_triple = def_out_triple_callback;

   if (-1 == create_sites (v))
     {
	voronoi_close (v);
	return NULL;
     }

   v->nvertices = v->nedges = 0 ;

   if (-1 == el_initialize (v))
     {
	voronoi_close (v);
	return NULL;
     }

   if (-1 == PQinitialize (v))
     {
	voronoi_close (v);
	return NULL;
     }

   return v;
}

typedef struct
{
   Voronoi_Triangulate_Type *triangles;
   unsigned int max_allocated_triangles;
   unsigned int delta_num;
}
Output_Triple_Type;

static int realloc_output_triples (Output_Triple_Type *t, unsigned int num)
{
   Voronoi_Triangulate_Type *tri;
   unsigned int i;

   tri = t->triangles;
   for (i = 0; i < 3; i++)
     {
	int *indx;
	if (NULL == (indx = (int *) SLrealloc ((char *)tri->vertex_indices[i], num*sizeof(int))))
	  return -1;
	tri->vertex_indices[i] = indx;
     }
   t->max_allocated_triangles = num;
   return 0;
}

static void free_output_triples (Output_Triple_Type *t)
{
   Voronoi_Triangulate_Type *tri;

   if (NULL == (tri = t->triangles))
     return;

   SLfree ((char *)tri->vertex_indices[0]);
   SLfree ((char *)tri->vertex_indices[1]);
   SLfree ((char *)tri->vertex_indices[2]);
}


static int out_triple_callback (Voronoi_Type *v, Site *s1, Site *s2, Site *s3)
{
   unsigned int num;
   Voronoi_Triangulate_Type *tri;
   Output_Triple_Type *t;

   if (v->triangulate == 0)
     return 0;

   t = (Output_Triple_Type *) v->out_triple_data;
   tri = t->triangles;
   num = tri->num_triangles;
   if (num == t->max_allocated_triangles)
     {
	if (-1 == realloc_output_triples (t, num + t->delta_num))
	  return -1;
     }

   tri->vertex_indices[0][num] = s1->sitenbr;
   tri->vertex_indices[1][num] = s2->sitenbr;
   tri->vertex_indices[2][num] = s3->sitenbr;

   tri->num_triangles = num+1;
   return 0;
}

int voronoi_triangulate (double *x, double *y, unsigned int npts,
			 Voronoi_Triangulate_Type *tri)
{
   Voronoi_Type *v;
   Output_Triple_Type t;

   memset ((char *) tri, 0, sizeof (Voronoi_Triangulate_Type));

   if (npts < 3)
     return 0;

   v = voronoi_open (x, y, npts);
   if (v == NULL)
     return -1;

   t.triangles = tri;
   t.max_allocated_triangles = 0;
   t.delta_num = 4 + npts / 10;

   if (-1 == realloc_output_triples (&t, 2*npts))
     goto return_error;

   v->out_triple = out_triple_callback;
   v->out_triple_data = &t;

   if (-1 == voronoi (v, 1))
     goto return_error;

   if (-1 == realloc_output_triples (&t, t.triangles->num_triangles))
     goto return_error;

   voronoi_close (v);
   return 0;

return_error:
   voronoi_close (v);
   free_output_triples (&t);
   return -1;
}

static double *Convex_Hull_X_Pts;
static double *Convex_Hull_Y_Pts;
static int lower_hull_cmp_func (const void *va, const void *vb)
{
   double diff;
   diff = Convex_Hull_X_Pts[*(int *)va] - Convex_Hull_X_Pts[*(int *)vb];
   if (diff > 0) return 1;
   if (diff < 0) return -1;
   diff = Convex_Hull_Y_Pts[*(int *)vb] - Convex_Hull_Y_Pts[*(int *)va];
   if (diff > 0) return 1;
   if (diff < 0) return -1;
   return 0;
}

static int upper_hull_cmp_func (const void *vb, const void *va)
{
   double diff;
   diff = Convex_Hull_X_Pts[*(int *)va] - Convex_Hull_X_Pts[*(int *)vb];
   if (diff > 0) return 1;
   if (diff < 0) return -1;
   diff = Convex_Hull_Y_Pts[*(int *)vb] - Convex_Hull_Y_Pts[*(int *)va];
   if (diff > 0) return 1;
   if (diff < 0) return -1;
   return 0;
}

static int is_ccw (double *x, double *y,
		   unsigned int i, unsigned int j, unsigned int k)
{
   double xj = x[j], yj = y[j];
   return ((x[i]-xj)*(y[k]-yj) - (y[i]-yj)*(x[k]-xj) <= 0);
}

static int make_chain (double *x, double *y, unsigned int num, int *indices,
		       int (*cmpfunc)(const void*, const void*))
{
   unsigned int i, s;

   Convex_Hull_X_Pts = x;
   Convex_Hull_Y_Pts = y;
   qsort (indices, num, sizeof(int), cmpfunc);

   s = 1;
   for (i = 2; i < num; i++)
     {
	int ii;
	unsigned int j = s;
	ii = indices[i];
	while ((j >= 1) && is_ccw(x, y, ii, indices[j], indices[j-1]))
	  j--;
	s = j + 1;

	indices[i] = indices[s]; indices[s] = ii;
     }
   return s;
}

int voronoi_convex_hull (double *x, double *y, unsigned int num_pts,
			 int **indicesp, unsigned int *num_indicesp)
{
   int *indices, *tmp_indices;
   unsigned int i, n;

   if (NULL == (indices = (int *)malloc (sizeof(int)*(num_pts+1))))
     return -1;

   for (i = 0; i < num_pts; i++) indices[i] = i;
   n = make_chain (x, y, num_pts, indices, lower_hull_cmp_func);
   if (n != 0)
     {
	indices[num_pts] = indices[0];
	n += make_chain (x, y, num_pts-n+1, indices+n, upper_hull_cmp_func);
     }
   tmp_indices = (int *) realloc (indices, sizeof(int)*((n == 0) ? 1 : n));
   if (tmp_indices == NULL)
     {
	free (indices);
	return -1;
     }
   *num_indicesp = n;
   *indicesp = tmp_indices;
   return 0;
}
